import linecache
import logging
import sys

import structlog  # type: ignore
from structlog.types import EventDict, WrappedLogger

# to clean out AWS Lambda logging configuration
root = logging.getLogger()
if root.handlers:
    for handler in root.handlers:
        root.removeHandler(handler)

logging.basicConfig(format="%(message)s",
                    stream=sys.stdout,
                    level=logging.INFO)


class StructuredExceptionRenderer:
    """
    Transform exc_info (if present) into a Dict and place into the `exception` key. This is similar to the behavior of
    the Java net.logstash.logback.argument.StructuredArguments package.
    """

    def __call__(self, logger: WrappedLogger, name: str, event_dict: EventDict) -> EventDict:
        if not event_dict.pop("exc_info", None):
            return event_dict

        # exception class, the exception, the traceback
        (type, value, tb) = sys.exc_info()

        if type is None:
            return event_dict

        stack = []
        while tb is not None:
            stack.append({
                "methodName": tb.tb_frame.f_code.co_name,
                "fileName": tb.tb_frame.f_code.co_filename,
                "lineNumber": tb.tb_lineno,  # first tb.tb_frame.f_lineno is of the logger.exception() call
                "lineText": linecache.getline(tb.tb_frame.f_code.co_filename, tb.tb_lineno).strip()
            })
            tb = tb.tb_next

        event_dict["exception"] = {
            "message": f"{type.__name__}: {str(value)}",
            "stackTrace": stack
        }

        return event_dict


structlog.configure_once(
    processors=[
        structlog.stdlib.filter_by_level,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.TimeStamper(fmt="iso", key="@timestamp"),
        structlog.processors.StackInfoRenderer(),
        StructuredExceptionRenderer(),  # instead of structlog.processors.format_exc_info
        # structlog.processors.format_exc_info,
        structlog.processors.UnicodeDecoder(),
        structlog.processors.JSONRenderer()
    ],
    context_class=dict,
    logger_factory=structlog.stdlib.LoggerFactory(),
    wrapper_class=structlog.stdlib.BoundLogger,
    cache_logger_on_first_use=True,
)

log = structlog.get_logger(__name__)
